import logging
import os
import decimal
import inspect
import boto3
from functools import wraps
from os.path import expanduser
import psutil, time
import numpy as np
from ec2_metadata import ec2_metadata
import json
import pytz
from datetime import datetime
from typing import NamedTuple

india_tz = pytz.timezone("Asia/Kolkata")

formatter = logging.Formatter(
    "%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s"
)


class ConfigKeys:
    use_firefox_for_kite_token = "use_firefox_for_kite_token"
    tradeview_username = "tradeview_username"
    tradeview_password = "tradeview_password"
    kite_api_key = "kite_api_key"
    kite_api_secret = "kite_api_secret"
    kite_username = "kite_username"
    kite_pin = "kite_pin"
    kite_password = "kite_password"
    kite_left_eye = "kite_left_eye"
    kite_fav_dish = "kite_fav_dish"
    kite_spouse_email = "kite_spouse_email"
    kite_tv = "kite_tv"
    kite_car_color = "kite_car_color"


class ConstString:
    base_path = (
        os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))).split(
            "AlphaPlayground"
        )[0]
        + "AlphaPlayground"
    )


class LoggerUtils:
    @staticmethod
    def get_log_file_path(file_name):
        today_date = datetime.now()

        return "/tmp/" + file_name + "-" + today_date.strftime("%Y-%m-%d:01") + ".log"

    @staticmethod
    def setup_logger(logger_name, log_file, level=logging.DEBUG):
        log_file = LoggerUtils.get_log_file_path(log_file)
        """Function setup as many loggers as you want"""

        handler = logging.FileHandler(log_file)
        handler.setFormatter(formatter)

        logger = logging.getLogger(logger_name)
        logger.setLevel(level)
        logger.addHandler(handler)
        logger.propagate = False
        logger.disabled = False

        return logger


class IntervalAttribute:
    seconds = "seconds"
    mod_value = "mod_value"


class TickInterval:
    """ToDo: Make Enum"""

    five_minute = "5minute"
    day = "day"
    thirty_minute = "30minute"
    three_minute = "3minute"
    minute = "minute"
    fifteen_minute = "15minute"

    @staticmethod
    def get_interval_seconds(interval: str) -> int:
        if interval == TickInterval.minute:
            return 60
        elif interval == TickInterval.five_minute:
            return 300
        elif interval == TickInterval.fifteen_minute:
            return 900
        elif interval == TickInterval.three_minute:
            return 180
        elif interval == TickInterval.day:
            return 375 * 60
        elif interval == TickInterval.thirty_minute:
            return 30 * 60
        else:
            raise Exception("Interval {} is not defined".format(interval))


class Constants:
    interval_attributes = {
        TickInterval.minute: {
            IntervalAttribute.seconds: 60,
            IntervalAttribute.mod_value: 1,
        },
        TickInterval.five_minute: {
            IntervalAttribute.seconds: 300,
            IntervalAttribute.mod_value: 5,
        },
        TickInterval.day: {
            IntervalAttribute.seconds: 375 * 60,
            IntervalAttribute.mod_value: 1,
        },
        TickInterval.thirty_minute: {IntervalAttribute.seconds: 30 * 60},
        TickInterval.three_minute: {
            IntervalAttribute.seconds: 180,
            IntervalAttribute.mod_value: 3,
        },
    }

    security_wise_delivery_position = "security_wise_delivery_position"
    daily_volatility = "daily-volatility"
    daily_indices = "daily-indices"
    open_interest = "open-interest"
    combined_open_interest = "combined-open-interest"
    fo_bhavcopy = "fo-bhavcopy"
    cm_bhavcopy = "cm-bhavcopy"
    execution_algorithm = "execution_algorithm"

    close_all_tags = "close_all_tags"
    not_tradable_stocks = []

    nifty_fifty = [
        "M&M",
        "TATASTEEL",
        "ULTRACEMCO",
        "BAJFINANCE",
        "HINDALCO",
        "VEDL",
        "JSWSTEEL",
        "EICHERMOT",
        "ADANIPORTS",
        "GRASIM",
        "CIPLA",
        "IOC",
        "UPL",
        "KOTAKBANK",
        "HDFC",
        "BPCL",
        "BAJAJFINSV",
        "HDFCBANK",
        "INFRATEL",
        "INFY",
        "INDUSINDBK",
        "TCS",
        "POWERGRID",
        "BRITANNIA",
        "ZEEL",
        "SBIN",
        "DRREDDY",
        "NTPC",
        "WIPRO",
        "HEROMOTOCO",
        "YESBANK",
        "HCLTECH",
        "SUNPHARMA",
        "ASIANPAINT",
        "HINDUNILVR",
        "TATAMOTORS",
        "MARUTI",
        "COALINDIA",
        "RELIANCE",
        "ITC",
        "AXISBANK",
        "TITAN",
        "BHARTIARTL",
        "LT",
        "ICICIBANK",
        "GAIL",
        "ONGC",
        "TECHM",
        "IBULHSGFIN",
    ]

    indexes_t = [
        "NIFTY 50",
        "NIFTY 100",
        "NIFTY 200",
        "NIFTY MIDCAP 50",
        "NIFTY MIDCAP 100",
        "INDIA VIX",
    ]
    indexes = ["NIFTY 50"]

    stocks_for_later = ["GRANULES"]

    number_of_buckets = 10
    stock_to_consider_for_trading = 400
    max_trade_amount = 50000
    trading_charges = 0.00101
    overnight_long_charges = 0.0003
    max_order_placement_amount = 50000
    max_single_order_amount = 75000
    candle_data_path = "/tmp/candle_data/"

    base_path = ConstString.base_path

    home = expanduser("~")

    stock_data_path = "{}/stock_data/".format(home)
    stock_temp_data = "{}/stock_temp_data/".format(home)

    logger = LoggerUtils.setup_logger("main_logger", "trading")
    order_execution_logger = LoggerUtils.setup_logger(
        "order_execution", "order_execution"
    )
    historical_data_logger = LoggerUtils.setup_logger(
        "hist_data_looger", "historical_data"
    )
    tick_logger = LoggerUtils.setup_logger("tick_logger", "tick")

    date_format_without_zone = "%Y-%m-%d %H:%M:%S"
    hour_min_format = "%H:%M"
    kite_object_path = "/tmp/kite_object.pickle"

    instance_type = ""
    region = ""
    try:
        region = ec2_metadata.region
        instance_type = ec2_metadata.instance_type
    except Exception as e:
        pass

    @staticmethod
    def move_to_local(data, stock_name, interval, folder_path=None):
        if folder_path == None:
            folder_path = Constants.base_path + "/tmp/stock_data_local/" + interval + "/"

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

        file_path = folder_path + stock_name + ".csv"

        print("saving stock at {}".format(file_path))
        data[stock_name].to_csv(file_path, sep=",", index=False)
        return

    @staticmethod
    def modification_date(filename):
        t = os.path.getmtime(filename)
        return datetime.fromtimestamp(t)

    @staticmethod
    def get_json(data):
        data = json.dumps(data, default=Constants.myconverter)
        data = json.loads(data)
        return data

    @staticmethod
    def myconverter(o):
        if isinstance(o, datetime):
            return o.__str__()
        if isinstance(o, decimal.Decimal):
            # wanted a simple yield str(o) in the next line,
            # but that would mean a yield on the line with super(...),
            # which wouldn't work (see my comment below), so...
            return float(o)
        if isinstance(o, np.int64):
            return int(o)

    @staticmethod
    def round_to_tick(number):
        return round(number * 20) / 20

    @staticmethod
    def get_india_datetime(d):
        timestamp = int(d.strftime("%s"))
        india_datetime = datetime.fromtimestamp(timestamp, india_tz)
        return india_datetime

    @staticmethod
    def getCurrentMemoryUsage():
        # gives a single float value
        print("CPU percentage {}".format(psutil.cpu_percent()))

        process = psutil.Process(os.getpid())
        mem = process.get_memory_info()[0] / float(2**20)
        print("Memory {}".format(mem))
        # you can convert that object to a dictionary

        return None

    @staticmethod
    def get_future_suffix(index):
        months = [
            "JAN",
            "FEB",
            "MAR",
            "APR",
            "MAY",
            "JUN",
            "JUL",
            "AUG",
            "SEP",
            "OCT",
            "NOV",
            "DEC",
        ]

        d = datetime.now()
        month = d.month
        return str(d.year)[-2:] + months[(month + index - 1) % 12] + "FUT"

    @staticmethod
    def get_alphas(day_candle_data, bucket="nifty", suffix="bucket", market_alpha=False):
        alphas = [
            k for k in day_candle_data.columns if "alpha" in k and "market_alpha" not in k
        ]

        if market_alpha == True:
            alphas = [k for k in day_candle_data.columns if "market_alpha" in k]

        if bucket == None:
            raise ("bucket can't be none")

        alphas = [k for k in alphas if bucket in k]
        alphas_copy = []
        for alpha in alphas:
            alphas_copy.append(alpha.split(bucket)[0] + bucket)

        alphas = list(set(alphas_copy))

        return [k + "_" + suffix for k in alphas]

    @staticmethod
    def get_futures_list(stocks, index=None):
        future_suffix = []

        if index == None:
            for i in range(0, 3):
                future_suffix.append(Constants.get_future_suffix(i))
        else:
            if index > 2 or index < 0:
                raise Exception("Invalid index value")
            future_suffix.append(Constants.get_future_suffix(index))

        response = []
        for stock in stocks:
            response.append(stock)

            if stock == "NIFTY 50":
                continue

            for f in future_suffix:
                response.append(stock + f)

        return response


class MarketTime(NamedTuple):
    hour: int
    minute: int

    def ticks_until(
        self, other: "MarketTime", tick_interval: str = TickInterval.five_minute
    ) -> int:
        """Returns number of ticks until the other market time where each tick covers tick_interval
        time. Tick interval defaults to TickInterval.five_minute

        Usage:
        ------
         >>> MarketTime(9,25).ticks_until(MarketTime(13,55))
         54
        """
        end_date = datetime.now().replace(hour=other.hour, minute=other.minute)
        start_date = end_date.replace(hour=self.hour, minute=self.minute)
        date_diff = end_date - start_date
        return int(
            date_diff.total_seconds() // TickInterval.get_interval_seconds(tick_interval)
        )


class SQSConstants:
    screen_snapshot_queue_url = (
        "https://sqs.ap-south-1.amazonaws.com/612481995471/screen_snapshot_queue"
    )


class ZerodhaServiceConstants:
    zerodha = "ZERODHA"
    token = "TOKEN"
    request_token = "request_token"
    kite_oject = "kite_oject"


class DynamoDbConstants:
    dynamodb = boto3.resource("dynamodb", region_name="ap-south-1")
    stock_table = dynamodb.Table("StockData")
    orders_table = dynamodb.Table("Orders")
    configuration_table = dynamodb.Table("Configuration")
    machine_task_table = dynamodb.Table("MachineTask")


class StrategyType:
    intraday = "intraday"
    interday = "interday"
    overnight = "overnight"
    overnight_long = "overnight_long"
    overnight_short = "overnight_short"


class ExchangeName:
    nse = "NSE"
    nfo = "NFO"


class StrategySegment:
    index = "index"
    future = "future"
    equity = "equity"


class StrategyBucket:
    future = "future"
    top_500 = "500"


class ExecutionAlgorithm:
    # Transaction type
    TRANSACTION_TYPE_BUY = "BUY"
    TRANSACTION_TYPE_SELL = "SELL"

    limit_order_20s_m = "limit_order_20s_m"
    limit_order_20s = "limit_order_20s"
    limit_order_20s_m2 = "limit_order_20s_m2"

    pike_place_market = "pike_place_market"
    dubrovnik_market = "dubrovnik_market"

    # M3 all modification for short sell orders will happen at buy price (including order placement)
    limit_order_20s_m3 = "limit_order_20s_m3"

    market = "market"


class S3Constants:
    historical_data = "chasing-alpha-historical-data"
    stock_data = "chasing-alpha-stock-data"


class S3Bucket:
    chasing_alpha_market_data = "chasing-alpha-market-data"
    chasing_alpha_ml_models_prod = "chasing-alpha-ml-models-prod"
    chasing_alpha_ml_models_beta = "chasing-alpha-ml-models-beta"


class StockTableColumns:
    AVG_VOL_90 = "AVG_VOL_90"
    Industry = "Industry"
    AVG_VOL_10 = "AVG_VOL_10"
    Last = "Last"
    StockPrice = "StockPrice"
    StockVolume = "StockVolume"
    P_95_5_MINUTE = "P_95_5_MINUTE"
    P_90_5_MINUTE = "P_90_5_MINUTE"
    P_70_5_MINUTE = "P_70_5_MINUTE"
    P_50_5_MINUTE = "P_50_5_MINUTE"
    P_60_5_MINUTE = "P_60_5_MINUTE"
    Volume = "Volume"
    Vwap = "VWAP"
    ZerodhaVwap = "ZerodhaVwap"
    TotalVolume = "TotalVolume"
    VwapPrice = "VwapPrice"
    stock = "stock"


class TickColumns:
    nifty_500_close = "nifty_500_close"
    quantityTraded = "quantityTraded"
    efficiency = "efficiency"


class StockVolume:
    Extreme = "Extreme"
    High = "High"
    Okay = "Okay"


class StockPrice:
    MegaCap = "MegaCap"
    Large = "Large"
    Penny = "Penny"


class DataframeFilters:
    @staticmethod
    def filter(df, apply_strategy=False):
        pass


class OrderType:
    market = "MARKET"
    slm = "SL-M"
    sl = "SL"
    limit = "LIMIT"


class Product:
    bracket = "BO"
    mis = "MIS"
    cnc = "CNC"
    nrml = "NRML"


class OrderVariety:
    bo = "bo"
    regular = "regular"


class PlatformArgs:
    push_trades = "push_trades"
    dump_day_candles = "dump_day_candles"


class Order:
    total_charges = "total_charges"
    stamp_charges = "stamp_charges"
    gst_charges = "gst_charges"
    brokerage_charges = "brokerage_charges"
    combined_exposure = "combined_exposure"
    exchange_transaction_charges = "exchange_transaction_charges"
    actual_tradingsymbol = "actual_tradingsymbol"
    order_type = "order_type"
    tag = "tag"
    disclosed_quantity = "disclosed_quantity"
    child_orders = "child_orders"
    quantity = "quantity"
    pending_quantity = "pending_quantity"
    last_price = "last_price"
    price = "price"
    strategy_price = "strategy_price"
    tradingsymbol = "tradingsymbol"
    exchange = "exchange"
    trade_id = "trade_id"
    product = "product"
    transaction_type = "transaction_type"
    order_timestamp = "order_timestamp"
    average_price = "average_price"
    entry_price = "entry_price"
    sell_side_price = "sell_side_price"
    buy_side_price = "buy_side_price"
    closing_timestamp = "closing_timestamp"
    squareoff = "squareoff"
    is_intraday = "is_intraday"
    stt_charges = "stt_charges"
    trade_amount = "trade_amount"
    trade_number = "trade_number"
    closeQuantity = "closeQuantity"
    close_quantity = "close_quantity"
    sell_transaction = "SELL"
    buy_transaction = "BUY"
    stoploss = "stoploss"
    oms_version = "oms_version"
    place_without_stoploss = "place_without_stoploss"
    pseudo_order_id = "pseudo_order_id"
    pseudo_parent_order_id = "pseudo_parent_order_id"
    close_price = "close_price"
    change = "change"
    seconds_since_midnight = "seconds_since_midnight"
    number_of_tick = "number_of_tick"
    mom = "mom"
    mfi = "mfi"
    traded_amount = "traded_amount"
    macd_width = "macd_width"
    bbwidth = "bbwidth"
    price_wrt_vwap = "price_wrt_vwap"
    rsi = "rsi"
    volatility = "volatility"
    percentage_change = "percentage_change"
    variety = "variety"
    profit = "profit"
    pnl_percentage = "pnl_percentage"
    exposure = "exposure"
    profit_without_transaction = "profit_without_transaction"
    status = "status"
    closed_by = "closed_by"
    buy = "BUY"
    sell = "SELL"
    day_trading_exit = "day_trading_exit"
    parent_order_id = "parent_order_id"
    child_orders = "child_orders"
    order_id = "order_id"
    trigger_price = "trigger_price"
    mode = "mode"
    is_position_closed = "is_position_closed"


class OrderTableColumns:
    order_id = Order.order_id
    execution_time = "execution_time"
    volumePercentile = "volumePercentile"
    order_price = "order_price"
    closed_by = "closed_by"
    balance = "balance"
    closingTimestamp = "closing_timestamp"
    execution_algorithm = "execution_algorithm"
    modification_count = "modification_count"
    child_modification_count = "child_modification_count"
    closeQuantity = "closeQuantity"
    close_price = Order.close_price
    mode = Order.mode
    weighted_intraday_close = "weighted_intraday_close"
    strategy_type = "strategy_type"
    strategy_bucket = "strategy_bucket"
    next_day_intraday_open = "next_day_intraday_open"

    all_columns = [
        Order.order_id,
        Order.order_timestamp,
        Order.closing_timestamp,
        Order.variety,
        Order.tradingsymbol,
        Order.transaction_type,
        Order.quantity,
        Order.profit,
        Order.tag,
        Order.average_price,
        execution_time,
        order_price,
        volumePercentile,
        closed_by,
        Order.entry_price,
        Order.exposure,
        Order.profit_without_transaction,
        balance,
        mode,
        Order.exposure,
        Order.close_price,
        Order.pnl_percentage,
        Order.last_price,
        modification_count,
        execution_algorithm,
        child_modification_count,
        strategy_type,
        strategy_bucket,
        Order.order_type,
        Order.trade_id,
        Order.place_without_stoploss,
        Order.exchange,
        Order.product,
        Order.is_position_closed,
    ]

    portfolio_columns = all_columns + [
        "symbol",
        "stoploss",
        "original_tag",
        weighted_intraday_close,
        next_day_intraday_open,
        "next_date",
    ]


class SlippageMetricsTableColumns:
    order_id = Order.order_id
    entry_slippage = "entry_slippage"
    exit_slippage = "exit_slippage"
    backtest_pnl = "backtest_pnl"
    order_timestamp = "order_timestamp"
    tt = "tt"
    original_pnl = "original_pnl"
    prod_entry_price = "prod_entry_price"
    prod_exit_price = "prod_exit_price"
    backtest_entry_price = "backtest_entry_price"
    backtest_exit_price = "backtest_exit_price"
    prod_exposure = "prod_exposure"
    backtest_exposure = "backtest_exposure"
    prod_pnl = "prod_pnl"
    prod_quantity = "prod_quantity"
    backtest_quantity = "backtest_quantity"

    all_columns = [
        order_id,
        backtest_pnl,
        Order.tag,
        Order.tradingsymbol,
        prod_entry_price,
        prod_exit_price,
        backtest_entry_price,
        backtest_exit_price,
        prod_pnl,
        order_timestamp,
        OrderTableColumns.execution_algorithm,
        prod_exposure,
        backtest_exposure,
        prod_quantity,
        backtest_quantity,
        OrderTableColumns.closingTimestamp,
        "liquidity_bucket_500",
        "liquidity",
        "liquidity_20",
        "is_fno",
    ]


class StrategyMetricsTableColumns:
    a_sharpe_ratio = "a_sharpe_ratio"
    weekly_sharpe_ratio = "weekly_sharpe_ratio"
    monthly_sharpe_ratio = "monthly_sharpe_ratio"
    monthly_pnl = "monthly_pnl"
    exposure_90 = "exposure_90"
    sharpe_ratio = "sharpe_ratio"
    winning_streak = "winning_streak"
    lossing_streak = "lossing_streak"
    avg_profit = "avg_profit"
    avg_loss = "avg_loss"
    max_drawdown = "max_drawdown"
    unique_stocks = "unique_stocks"
    max_drawdown_days = "max_drawdown_days"
    max_gain = "max_gain"
    max_loss = "max_loss"
    tag = "tag"
    trade_amount = "trade_amount"
    year = "year"
    avg_pnl = "avg_pnl"
    success_ratio = "success_ratio"
    bps = "bps"
    slippage = "slippage"
    total_signals = "total_signals"
    number_of_losses = "number_of_losses"
    number_of_wins = "number_of_wins"
    all_columns = [
        avg_pnl,
        success_ratio,
        total_signals,
        number_of_losses,
        number_of_wins,
        a_sharpe_ratio,
        sharpe_ratio,
        winning_streak,
        lossing_streak,
        avg_profit,
        avg_loss,
        max_drawdown,
        max_gain,
        max_loss,
        tag,
        year,
        weekly_sharpe_ratio,
        trade_amount,
        monthly_sharpe_ratio,
        bps,
        slippage,
        monthly_pnl,
        exposure_90,
        max_drawdown_days,
        unique_stocks,
    ]


class StrategyTags:
    vyakul = "vyakul"
    roger = "roger"
    messi = "messi"
    ronaldo = "ronaldo"
    main = "main"
    suarez = "suarez"
    amazon = "amazon"
    arham = "arham"
    bruce = "bruce"
    harvey = "harvey"
    jessica = "jessica"
    mario = "mario"
    kipchoge = "kipchoge"
    sam = "sam"
    ellen = "ellen"
    root = "root"
    murdock = "murdock"
    winry = "winry"
    jack = "jack"
    zlatan = "zlatan"
    shifu = "shifu"
    batman = "batman"
    djokovic = "djokovic"
    nadal = "nadal"
    python = "python"
    panda = "panda"
    kenmiles = "kenmiles"
    morningreversion = "morningreversion"


class Strategies:
    TestStrategy = "TestStrategy"
    Midday = "Midday"
    KenmilesN = "KenmilesN"
    Suarez = "Suarez"
    VyakulN = "VyakulN"
    Witcher = "Witcher"
    Jack = "Jack"
    Root = "Root"
    Finch = "Finch"
    Queen = "Queen"
    Harvey = "Harvey"
    Terstegen = "Terstegen"
    PortfolioN = "PortfolioN"
    MustangN = "MustangN"
    Stark = "Stark"
    Roach = "Roach"
    Musk = "Musk"
    Ellen = "Ellen"
    Trevor = "Trevor"
    Keanu = "Keanu"
    MorningReversion = "MorningReversion"
    House = "House"
    Roger = "Roger"
    Nadal = "Nadal"
    Matsya = "Matsya"
    Bruno = "Bruno"
    Cartman = "Cartman"
    Fnoshort = "Fnoshort"
    fno_port = "fno_port"
    terstegen_fno = "terstegen_fno"
    harvey_fno = "harvey_fno"


class OrderStatus:
    complete = "COMPLETE"
    cancelled = "CANCELLED"
    open = "OPEN"
    rejected = "REJECTED"
    trigger_pending = "TRIGGER PENDING"
    modify_validation_pending = "MODIFY VALIDATION PENDING"
    validation_pending = "VALIDATION PENDING"


class StrategyConfigs:
    strategy_config_path = "/quantplay/strategies/strategy_configs"
    base_path = Constants.base_path + strategy_config_path
    profiler = base_path + "/profiler.json"
    prod = base_path + "/prod.json"
    pre_prod = base_path + "/pre_prod.json"
    backtest_set1 = base_path + "/backtest_set1.json"


class TradingMode:
    replayTick = "replayTick"
    prod = "prod"
    backtest = "backtest"
    profiler = "profiler"
    prod_verfication = "prod_verification"
    optimize = "optimize"


class FeedName:
    S3 = "s3"


class MLModels:
    matsya_beta = "/tmp/matsya.sav"


class TradingCode:
    ok = "200"
    exit_all = "400"


def timeit(*args_main, **kwargs_main):
    def inner_function(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            ts = time.time()
            result = function(*args, **kwargs)
            te = time.time()
            if "MetricName" in kwargs_main:
                data = {
                    "MetricName": kwargs_main["MetricName"],
                    "Unit": "Seconds",
                    "Value": (te - ts),
                    "Dimensions": [
                        {"Name": "Region", "Value": Constants.region},
                        {
                            "Name": "InstanceType",
                            "Value": Constants.instance_type,
                        },
                    ],
                }
                Constants.logger.info("{}".format(json.dumps(data)))
            else:
                Constants.logger.error("No metric found in {}".format(function.__name__))
            return result

        return wrapper

    return inner_function



from quantplay.utils.constant import Order, StrategyType, ExchangeName
import numpy as np
import pandas as pd


class TransactionCost:

    ZERODHA_BROKERAGE_CHARGES = 20

    @staticmethod
    def add_transaction_costs(orders_df):
        """
        https://zerodha.com/charges/#tab-equities
        """
        orders_df.loc[:, Order.order_timestamp] = pd.to_datetime(
            orders_df.order_timestamp
        )
        orders_df.loc[:, Order.closing_timestamp] = pd.to_datetime(
            orders_df.closing_timestamp
        )

        orders_df.loc[:, Order.is_intraday] = np.where(
            orders_df.strategy_type == StrategyType.intraday,
            True,
            False,
        )

        orders_df.loc[:, Order.sell_side_price] = np.where(
            orders_df[Order.transaction_type] == Order.sell_transaction,
            orders_df[Order.entry_price],
            orders_df[Order.close_price],
        )
        orders_df.loc[:, Order.buy_side_price] = np.where(
            orders_df[Order.transaction_type] == Order.buy_transaction,
            orders_df[Order.entry_price],
            orders_df[Order.close_price],
        )

        orders_df.loc[:, Order.combined_exposure] = (
            orders_df[Order.sell_side_price] + orders_df[Order.buy_side_price]
        ) * orders_df[Order.quantity]

        if "exchange" not in orders_df.columns:
            orders_df.loc[:, "exchange"] = ExchangeName.nse

        charges_condition = [
            (orders_df[Order.is_intraday] == True)
            & (orders_df["exchange"] == ExchangeName.nse),
            (orders_df[Order.is_intraday] == False)
            & (orders_df["exchange"] == ExchangeName.nse),
            (orders_df["exchange"] == ExchangeName.nfo)
            & (orders_df["security_type"] == "FUT"),
            (orders_df["exchange"] == ExchangeName.nfo)
            & (orders_df["security_type"] == "OPT"),
        ]

        brokerage_charges_choices = [
            0.0003 * orders_df[Order.combined_exposure],
            0,
            40,
            40,
        ]

        orders_df.loc[:, Order.brokerage_charges] = np.select(
            charges_condition, brokerage_charges_choices, default=0
        )

        stt_charges_choices = [
            0.00025 * (orders_df[Order.sell_side_price] * orders_df[Order.quantity]),
            0.001 * orders_df[Order.combined_exposure],
            0.0001 * (orders_df[Order.sell_side_price] * orders_df[Order.quantity]),
            0.0005 * (orders_df[Order.sell_side_price] * orders_df[Order.quantity]),
        ]

        orders_df.loc[:, Order.stt_charges] = np.select(
            charges_condition, stt_charges_choices, default=0
        )

        exchange_charges_choices = [
            0.0000345 * orders_df[Order.combined_exposure],
            0.0000345 * orders_df[Order.combined_exposure],
            0.00002 * orders_df[Order.combined_exposure],
            0.00053 * orders_df[Order.combined_exposure],
        ]

        orders_df.loc[:, Order.exchange_transaction_charges] = np.select(
            charges_condition, exchange_charges_choices, default=0
        )

        stamp_charges_choices = [
            0.00003 * orders_df[Order.buy_side_price] * orders_df[Order.quantity],
            0.00015 * orders_df[Order.buy_side_price] * orders_df[Order.quantity],
            0.00002 * orders_df[Order.buy_side_price] * orders_df[Order.quantity],
            0.00003 * orders_df[Order.buy_side_price] * orders_df[Order.quantity],
        ]

        orders_df.loc[:, Order.stamp_charges] = np.select(
            charges_condition, stamp_charges_choices, default=0
        )

        orders_df.loc[:, Order.gst_charges] = (
            0.18 * orders_df[Order.exchange_transaction_charges]
        )

        orders_df.loc[:, Order.total_charges] = (
            orders_df[Order.stt_charges]
            + orders_df[Order.exchange_transaction_charges]
            + orders_df[Order.stamp_charges]
            + orders_df[Order.gst_charges]
            + orders_df[Order.brokerage_charges]
        )

        return orders_df



from quantplay.order_execution.execution_algorithm import (
    OrderExecutionAlgorithm,
)
from quantplay.utils.constant import Constants


class MeanPriceExecutionAlgo(OrderExecutionAlgorithm):
    def __init__(self, max_modification_count, *args, **kwargs):
        self.name = "MeanPriceExecutionAlgo"
        super(MeanPriceExecutionAlgo, self).__init__(max_modification_count)

    def get_new_price(
        self,
        last_tick,
        quantplay_order,
        is_new_order=False,
        buy_side_factor=1,
        sell_side_factor=1,
    ):

        new_price = None

        if is_new_order:
            if quantplay_order.price:
                new_price = quantplay_order.price
            else:
                new_price = last_tick.last_price
        else:
            order_book = last_tick.order_book
            sell_side, buy_side = order_book.sell_side, order_book.buy_side
            new_price = (sell_side[0].price + buy_side[0].price) / 2

        if quantplay_order.transaction_type == "SELL":
            new_price = new_price * sell_side_factor
        else:
            new_price = new_price * buy_side_factor

        return Constants.round_to_tick(new_price)

    def get_next_modification_time(self, last_tick, quantplay_order):
        return 10



import codecs
import getpass
import json
import pickle

import numpy as np
import pandas as pd
import requests
import time
from smartapi import SmartConnect
from smartapi import SmartWebSocket

from retrying import retry
from quantplay.broker.generics.broker import Broker
from quantplay.config.qplay_config import QplayConfig
from quantplay.model.exchange.instrument import QuantplayInstrument
from quantplay.utils.constant import Constants
from quantplay.exception.exceptions import InvalidArgumentException
from quantplay.utils.exchange import Market as MarketConstants

class AngelOne(Broker):
    angelone_api_key = "angelone_api_key"
    angelone_api_secret = "angelone_api_secret"
    angelone_client_id = "angelone_client_id"
    angelone_wrapper = "angelone_wrapper"
    angel_refresh_token = "angel_refresh_token"

    def __init__(self):
        try:
            wrapper = QplayConfig.get_value(AngelOne.angelone_wrapper)
            self.wrapper = pickle.loads(codecs.decode(wrapper.encode(), "base64"))
            self.refreshToken = QplayConfig.get_value(AngelOne.angel_refresh_token)
            user_profile_response = self.wrapper.getProfile(self.refreshToken)
            if user_profile_response['message'] != "SUCCESS":
                raise Exception("AngelOne Token Expired")
            else:
                Constants.logger.info(user_profile_response)
        except Exception as e:
            Constants.logger.error(e)
            self.wrapper = self.generate_token()
            Constants.logger.info(self.wrapper.getProfile(self.refreshToken))
        self.refreshToken = QplayConfig.get_value(AngelOne.angel_refresh_token)
        self.client_id = QplayConfig.get_value(AngelOne.angelone_client_id)

        self.angelone_ws = SmartWebSocket(self.wrapper.getfeedToken(), self.client_id)

        super(AngelOne, self).__init__()
        self.populate_instruments()

    def get_symbol_token(self, exchange, tradingsymbol):
        return self.exchange_symbol_to_instrument_id_map[exchange][tradingsymbol]

    @retry(wait_exponential_multiplier=3000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def get_ltp(self, exchange=None, tradingsymbol=None):
        if tradingsymbol in MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP:
            tradingsymbol = MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP[tradingsymbol]

        symboltoken = self.get_symbol_token(exchange, tradingsymbol)

        if exchange == "NSE" and tradingsymbol not in ["NIFTY", "BANKNIFTY"]:
            tradingsymbol = "{}-EQ".format(tradingsymbol)

        response = self.wrapper.ltpData(exchange, tradingsymbol, symboltoken)
        if 'status' in response and response['status'] == False:
            raise InvalidArgumentException("Failed to fetch ltp broker error {}".format(response))

        return response['data']['ltp']

    def place_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                    tag=None, product=None, price=None, trigger_price=None):
        order = {}
        order["transactiontype"] = transaction_type

        order["variety"] = "NORMAL"
        order["tradingsymbol"] = tradingsymbol
        if order_type == "SL":
            order["variety"] = "STOPLOSS"

        if exchange == "NSE":
            order["tradingsymbol"] = "{}-EQ".format(tradingsymbol)

        order["triggerprice"] = trigger_price
        order["exchange"] = exchange
        order['symboltoken'] = self.get_symbol_token(exchange, tradingsymbol)

        if order_type == "SL":
            order_type = "STOPLOSS_LIMIT"
        order['ordertype'] = order_type

        if product == "MIS":
            product = "INTRADAY"
        elif product == "NRML":
            product = "CARRYFORWARD"
        order['producttype'] = product

        order['price'] = price
        order['quantity'] = quantity
        order["duration"] = "DAY"

        try:
            print("Placing order {}".format(json.dumps(order)))
            return self.wrapper.placeOrder(order)
        except Exception as e:
            exception_message = "Order placement failed with error [{}]".format(str(e))
            print(exception_message)

    @retry(wait_exponential_multiplier=3000, wait_exponential_max=10000, stop_max_attempt_number=3)
    def get_orders(self):
        order_book = self.wrapper.orderBook()
        if order_book['data']:
            return order_book['data']
        else:
            if 'errorcode' in order_book and order_book['errorcode'] == "AB1010":
                Constants.logger.error(
                    "Can't Fetch order book because session got expired")
            else:
                Constants.logger.error(
                    "Unknow error while fetching order book [{}]".format(order_book))

    def modify_orders_till_complete(self, orders_placed):
        modification_count = {}
        while 1:
            time.sleep(10)
            orders = pd.DataFrame(self.get_orders())

            orders = orders[orders.orderid.isin(orders_placed)]
            orders = orders[~orders.orderstatus.isin(["rejected", "cancelled", "complete"])]

            if len(orders) == 0:
                Constants.logger.info("ALL orders have be completed")
                break

            orders = orders.to_dict('records')
            for order in orders:
                order_id = order['orderid']

                ltp = self.get_ltp_by_order(order)
                order['price'] = ltp
                self.modify_order(order)

                if order_id not in modification_count:
                    modification_count[order_id] = 1
                else:
                    modification_count[order_id] += 1

                time.sleep(.1)

                if modification_count[order_id] > 5:
                    order['ordertype'] = "MARKET"
                    order['price'] = 0
                    Constants.logger.info("Placing MARKET order [{}]".format(order))
                    self.modify_order(order)

    def populate_instruments(self):
        Constants.logger.info("Setting up AngelOne client")
        data = requests.get("https://margincalculator.angelbroking.com/OpenAPI_File/files/OpenAPIScripMaster.json")
        inst_data = json.loads(data.content)
        inst_data = pd.DataFrame(inst_data)
        inst_data.loc[:, 'exchange'] = inst_data.exch_seg
        inst_data = inst_data[inst_data.exchange.isin(["NSE", "NFO"])]
        inst_data.loc[:, 'instrument_token'] = inst_data.token.astype(int)
        inst_data.loc[:, 'symbol'] = inst_data['symbol'].str.replace('-EQ','')

        assert set(['OPTSTK', 'OPTIDX', 'FUTSTK', 'FUTIDX']) == set(inst_data[inst_data.exch_seg == "NFO"].instrumenttype.unique())

        inst_data.loc[:, 'segment'] = None
        inst_data.loc[:, 'segment'] = np.where((inst_data.exch_seg == "NFO") & (
                    (inst_data.instrumenttype == "OPTIDX") | (inst_data.instrumenttype == "OPTSTK")),
                                               "NFO-OPT", inst_data.segment)
        inst_data.loc[:, 'segment'] = np.where((inst_data.exch_seg == "NFO") & (
                    (inst_data.instrumenttype == "FUTIDX") | (inst_data.instrumenttype == "FUTSTK")),
                                               "NFO-FUT", inst_data.segment)
        inst_data.loc[:, 'segment'] = np.where(inst_data.exch_seg == "NSE",
                                               "NSE", inst_data.segment)
        inst_data = inst_data[~inst_data.segment.isna()]
        inst_data.loc[:, 'instrument_type'] = np.where(inst_data.segment == "NFO-FUT", "FUT", None)
        inst_data.loc[:, 'instrument_type'] = np.where(inst_data.segment == "NSE", "EQ",
                                                       inst_data.instrument_type)
        inst_data.loc[:, 'instrument_type'] = np.where(
            ((inst_data.segment == "NFO-OPT") & (inst_data.symbol.str[-2:] == "PE")),
            "PE", inst_data.instrument_type)
        inst_data.loc[:, 'instrument_type'] = np.where(
            ((inst_data.segment == "NFO-OPT") & (inst_data.symbol.str[-2:] == "CE")),
            "CE", inst_data.instrument_type)
        inst_data.to_csv('/tmp/inst_data.csv')
        inst_data = inst_data.to_dict('records')

        instruments = list(
            map(
                lambda z_instrument: QuantplayInstrument.from_angelone_instrument(
                    z_instrument
                ),
                inst_data,
            )
        )
        self.instruments = instruments

        Broker.populate_instruments(self, instruments)


    def configure(self):
        quantplay_config = QplayConfig.get_config()

        print("Enter AngelOne API key:")
        api_key = input()

        print("Enter AngelOne API Secret:")
        api_secret = input()

        print("Enter AngelOne Client ID:")
        client_id = input()

        quantplay_config['DEFAULT'][AngelOne.angelone_api_key] = api_key
        quantplay_config['DEFAULT'][AngelOne.angelone_api_secret] = api_secret
        quantplay_config['DEFAULT'][AngelOne.angelone_client_id] = client_id

        with open('{}/config'.format(QplayConfig.config_path), 'w') as configfile:
            quantplay_config.write(configfile)

    def validate_config(self, quantplay_config):
        if quantplay_config is None:
            return False
        if AngelOne.angelone_api_key not in quantplay_config['DEFAULT']:
            return False
        if AngelOne.angelone_api_secret not in quantplay_config["DEFAULT"]:
            return False
        if AngelOne.angelone_client_id not in quantplay_config["DEFAULT"]:
            return False

        return True

    def option_symbol(self, underlying_symbol, expiry_date, strike_price, type):
        option_symbol = MarketConstants.INDEX_SYMBOL_TO_DERIVATIVE_SYMBOL_MAP[underlying_symbol]
        option_symbol += expiry_date.strftime("%d")
        option_symbol += expiry_date.strftime("%b").upper()
        option_symbol += expiry_date.strftime('%y')

        option_symbol += str(int(strike_price))
        option_symbol += type

        return option_symbol

    def generate_token(self):
        quantplay_config = QplayConfig.get_config()

        if not self.validate_config(quantplay_config):
            self.configure()
            quantplay_config = QplayConfig.get_config()
        api_key = quantplay_config['DEFAULT'][AngelOne.angelone_api_key]
        api_secret = quantplay_config['DEFAULT'][AngelOne.angelone_api_secret]
        client_id = quantplay_config['DEFAULT'][AngelOne.angelone_client_id]
        wrapper = SmartConnect(api_key=api_key)

        password = getpass.getpass()
        data = wrapper.generateSession(client_id, password)

        if 'message' in data and 'status' in data and data['status'] == False:
            print(data['message'])
            raise Exception("Token generation Failed")

        self.refreshToken = data['data']['refreshToken']
        QplayConfig.save_config(AngelOne.angel_refresh_token, self.refreshToken)

        QplayConfig.save_config("angelone_wrapper", codecs.encode(pickle.dumps(wrapper), "base64").decode())
        return wrapper


from typing import NamedTuple


class QuantplayInstrument(NamedTuple):
    """Common instrument model to be used across
    different exchanges and brokers.
    """

    symbol: str
    instrument_id: str
    instrument_type: str  # e.g EQ, CE, PE, FUT
    exchange: str  # e.g. NSE, NFO
    segment: str = ""  # e.g. NSE, NFO-OPT, NFO-FUT

    def __repr__(self):
        return "{} {} {} {} {}".format(
            self.symbol,
            self.instrument_id,
            self.instrument_type,
            self.exchange,
            self.segment,
        )

    def security_type(self):
        if self.instrument_type in ["CE", "PE"]:
            return "OPT"
        return self.instrument_type

    @classmethod
    def from_zerodha_instrument(cls, zerodha_instrument_dict):
        """Creates QuantplayInstrument from instrument dict returned by Zerodha API"""
        d = zerodha_instrument_dict
        return cls(
            d["tradingsymbol"],
            d["instrument_token"],
            d["instrument_type"],
            d["exchange"],
            d["segment"],
        )

    @classmethod
    def from_angelone_instrument(cls, angelone_instrument_dict):
        """Creates QuantplayInstrument from instrument dict returned by AngelOne API"""
        d = angelone_instrument_dict
        return cls(
            d["symbol"],
            d["instrument_token"],
            d["instrument_type"],
            d["exchange"],
            d["segment"],
        )



import re
from path import Path
from setuptools import setup, find_packages


def is_requirement(s: str) -> bool:
    """Returns true if the string is a valid requirement.
    For the requirement to be valid, it must be non empty
    and must not start from - or #.
    """
    s = s.strip()
    return not any([not s, re.match("^[-#]", s)])  # empty  # option or comment


requirements = [
    line.strip()
    for line in Path("requirements.txt").read_text().splitlines()
    if is_requirement(line)
]

setup(
    name="quantplay",
    long_description=Path("README.md").read_text(),
    version="1.1.2",
    setup_requires=["pytest-runner"],
    install_requires=requirements,
    tests_require=[],
    packages=find_packages(),
    url="",
    license="MIT",
    author="",
    author_email="",
    description="This python package will be stored in AWS CodeArtifact",
)



import configparser
from os.path import expanduser
import os


class QuantplayConfig:
    """Commands for easy backtesting strategy"""

    config_path = "{}/.quantplay".format(expanduser("~"))

    def __init__(self):
        pass

    @staticmethod
    def get_credentials():
        isExist = os.path.exists(QuantplayConfig.config_path)
        if not isExist:
            # Create a new directory because it does not exist
            os.makedirs(QuantplayConfig.config_path)

        credentials = configparser.ConfigParser()
        credentials.read("{}/config".format(QuantplayConfig.config_path))

        return credentials

    @staticmethod
    def get_config():
        return QuantplayConfig.get_credentials()



from quantplay.utils.constant import OrderType


class OrderExecutionAlgorithm:
    """Base class to be extended by different execution algorithms
    which can be assigned to different orders and easily A/B tested.
    """

    def __init__(self, max_modification_count, *args, **kwargs):
        """Creates an object of OrderExecutionAlgorithm. When
        max_modification_count number of modifications are done
        to the order, the order is converted to the Market order.

        :param: max_modification_count: int, between [1, 15]
        """

        if (
            not isinstance(max_modification_count, int)
            or max_modification_count < 1
            or max_modification_count > 15
        ):
            raise Exception(
                "Must provide max_modification_count as an int between 1 and 15"
            )

        self.max_modification_count = max_modification_count

    def modify_order(self, last_tick, quantplay_order, is_new_order=False):
        """Modifies the quantplay_order in_place before sending to the
        exchange
        """

        # if quantplay_order.modifications == self.max_modification_count:
        #     quantplay_order.order_type = str(OrderType.market)
        #     return
        quantplay_order.order_type = str(OrderType.market)
        quantplay_order.price = self.get_new_price(
            last_tick, quantplay_order, is_new_order=is_new_order
        )

        quantplay_order.modifications += 1

    def get_new_price(self, last_tick, quantplay_order, is_new_order=False, **kwargs):
        """Sets the price in the quantplay order
        for the next modification
        """
        raise NotImplementedError

    def get_next_modification_time(self, last_tick, quantplay_order):
        """Returns time in seconds after which the quantplay
        order should be modified
        """
        raise NotImplementedError

    def __repr__(self):
        return str(
            {"name": self.name, "max_modification_count": self.max_modification_count}
        )



import glob
import pandas as pd
import numpy as np
import copy
import time
import threading
from collections import defaultdict
from datetime import datetime, timedelta
from quantplay.utils.constant import Constants, IntervalAttribute, TickInterval
from quantplay.model.exchange.tick import QuantplayTick
from quantplay.utils.exchange import Market
from quantplay.model.exchange.order import (
    QuantplayExchangeResponse,
    QuantplayExchangeResponseType,
    QuantplayTransactionStatus,
    QuantplayExchangeOrderStatus,
)
from quantplay.exception.exceptions import QuantplayOrderPlacementException
from queue import Queue, PriorityQueue
import traceback


class Broker:
    """Parent class for all brokerages."""

    def __init__(self, tick_intervals, live_trade, *args, **kwargs):
        """Initiates the broker object to be used for fetching either
        a) historical data, OR
        b) live trading

        :param: tick_intervals: list of tick_interval
        :param: live_trade: bool, set to True if the broker object is to
        be used for live trade
        """
        self.tick_intervals = tick_intervals
        self.live_trade = live_trade
        self.strategy_instruments = (
            []
        )  # Dataframe will be created for only these instruments. Subset of streaming_instruments
        self.streaming_instruments = (
            []
        )  # Data will be streamed for all of these instruments
        self.instrument_id_to_symbol_map = dict()
        self.instrument_id_to_exchange_map = dict()
        self.instrument_id_to_security_type_map = dict()
        self.exchange_symbol_to_instrument_id_map = defaultdict(dict)

        # Live Market Feed Members
        self.exchange_symbol_last_tick = defaultdict(dict)
        self.pre_market_volume = defaultdict(int)
        self.live_market_df_by_interval = {
            interval: pd.DataFrame() for interval in self.tick_intervals
        }
        self.ongoing_tick_interval_time = dict()
        self.ongoing_tick_updates = dict()
        self.last_updated_interval_tick = dict()
        self.prev_tick_volume = dict()
        self.processed_tick_interval_queue = Queue()

        # Order Execution Members
        self.strategy_child_orders = defaultdict(
            list
        )  # {strategy_tag: child_quantplay_order}
        self.last_processed_order_map = dict()  # {order_id: quantplay_order}
        self.in_progress_order_map = dict()  # {order_id: quantplay_order}
        self.next_modifications_queue = PriorityQueue()

        # map locks
        self.order_placement_lock = threading.Lock()

        for tick_interval in self.tick_intervals:
            self.ongoing_tick_updates[tick_interval] = {}
            self.ongoing_tick_interval_time[tick_interval] = None
            self.last_updated_interval_tick[tick_interval] = None
            self.prev_tick_volume[tick_interval] = {}

    def load_historical_market_feed(self, days):
        instrument_data_by_interval = defaultdict(list)

        for instrument_id in self.strategy_instruments:
            symbol = self.instrument_id_to_symbol_map[instrument_id]
            security_type = self.instrument_id_to_security_type_map[instrument_id]

            end_date = datetime.now()
            end_date = end_date.replace(second=0, microsecond=0)

            start_date = end_date - timedelta(days=days)
            start_date = start_date.replace(minute=0, hour=0)

            response = self.get_historical_data(
                instrument_id, start_date=start_date, end_date=end_date
            )
            for interval, data in response.items():
                if len(data) > 0:
                    data = self.add_today_candles(data, symbol)
                    data.loc[:, "security_type"] = security_type
                    instrument_data_by_interval[interval].append(data)
                else:
                    Constants.logger.info(
                        "No historical data for symbol {} instrument_id {} interval {}".format(
                            symbol, instrument_id, interval
                        )
                    )

        for interval, df_list in instrument_data_by_interval.items():
            self.live_market_df_by_interval[interval] = pd.concat(df_list).reset_index(
                drop=True
            )

        self.reset_market_df_index()
        for interval in self.live_market_df_by_interval:
            self.live_market_df_by_interval["minute"].to_csv(
                f"/tmp/live_feed_{interval}.csv"
            )

    def add_today_candles(self, data, symbol):
        """Adds todays empty candles for the given interval to the data"""
        current_time = datetime.now()
        unique_times = data.date.dt.time.sort_values().unique()
        last_candle_time = data.iloc[-1].date.time()

        candles_to_add = []
        for candle_time in unique_times:
            if (data.iloc[-1].date.date() < datetime.now().date()) or (
                candle_time > last_candle_time
            ):
                candles_to_add.append(
                    {
                        "date": current_time.replace(
                            hour=candle_time.hour,
                            minute=candle_time.minute,
                            microsecond=0,
                            second=0,
                        ),
                        "symbol": symbol,
                    }
                )

        candles_to_add_df = pd.DataFrame(candles_to_add)
        return pd.concat([data, candles_to_add_df])

    def populate_instruments(self, instruments):
        """Fetches instruments for all exchanges from the broker
        and stores them in the member attributes.
        """
        Constants.logger.info("populating instruments")
        for instrument in instruments:
            exchange, symbol, instrument_id = (
                instrument.exchange,
                instrument.symbol,
                instrument.instrument_id,
            )
            self.instrument_id_to_symbol_map[instrument_id] = symbol
            self.instrument_id_to_exchange_map[instrument_id] = exchange
            self.instrument_id_to_security_type_map[
                instrument_id
            ] = instrument.security_type()
            self.exchange_symbol_to_instrument_id_map[exchange][symbol] = instrument_id


    def get_all_child_orders(self, *args, **kwargs):
        """Returns list of child orders by calling broker API"""
        pass

    def bootstrap_strategy_child_orders_map(self, strategies):
        """Bootstraps broker with child orders.

        :param: strategies: List[strategy objects]
        """
        Constants.logger.info("Bootstrapping child orders map")
        tag_to_strategy_map = {s.strategy_tag: s for s in strategies}
        child_orders = self.get_all_child_orders()
        self.order_placement_lock.acquire()
        for order in child_orders:
            strategy = tag_to_strategy_map[order.tag]
            order.set_strategy_attributes(strategy)
            self.strategy_child_orders[order.tag].append(order)
            self.in_progress_order_map[order.order_id] = order
            Constants.logger.info(f"Added Child Order {order}")
        self.order_placement_lock.release()
        Constants.logger.info(f"Added {len(child_orders)} total child orders")

    def add_strategy_symbols(self, symbols, exchange):
        """Adds symbols for which either
        a) historical data needs to be fetched, or
        b) live market feed needs to be streamed

        It updates self.strategy_instruments with the instruments
        matching the given symbols and exchange
        """
        tradable_symbols = [
            self.exchange_symbol_to_instrument_id_map[exchange][symbol]
            for symbol in symbols
            if symbol in self.exchange_symbol_to_instrument_id_map[exchange]
        ]
        self.strategy_instruments += tradable_symbols

        not_found_symbols = [symbol for symbol in symbols
            if symbol not in self.exchange_symbol_to_instrument_id_map[exchange]]

        Constants.logger.error("Symbols [{}] not found in data provided by broker".format(not_found_symbols))
        self.streaming_instruments = list(
            set(self.streaming_instruments + self.strategy_instruments)
        )
        print("Adding {} {} symbols for streaming".format(len(tradable_symbols), exchange))

    def stream_new_symbols(self, instrument_ids):
        """Adds instrument_ids for which data needs to be streamed.

        Note that live_market_df_by_interval will not contain data for
        these instrument_ids. Only exchange_symbol_last_tick will have
        data for these instrument_ids
        """
        pass

    def connect(self):
        """To be used for final connect after assigning callbacks"""
        pass

    def is_valid_live_tick(self, tick):
        """Returns True if the tick is valid"""
        raise NotImplementedError

    def __upsert_ongoing_tick(
        self,
        interval,
        symbol,
        instrument_id,
        last_price,
        volume_till_now,
        initialize=False,
    ):
        if initialize:
            self.ongoing_tick_updates[interval][symbol] = {}
            self.ongoing_tick_updates[interval][symbol]["open"] = last_price
            self.ongoing_tick_updates[interval][symbol]["high"] = last_price
            self.ongoing_tick_updates[interval][symbol]["low"] = last_price
            self.ongoing_tick_updates[interval][symbol]["symbol"] = symbol
            self.ongoing_tick_updates[interval][symbol][
                "security_type"
            ] = self.instrument_id_to_security_type_map[instrument_id]

        self.ongoing_tick_updates[interval][symbol]["high"] = max(
            last_price, self.ongoing_tick_updates[interval][symbol]["high"]
        )
        self.ongoing_tick_updates[interval][symbol]["low"] = min(
            last_price, self.ongoing_tick_updates[interval][symbol]["low"]
        )
        self.ongoing_tick_updates[interval][symbol]["close"] = last_price
        self.ongoing_tick_updates[interval][symbol]["volume"] = volume_till_now
        self.ongoing_tick_updates[interval][symbol]["total_volume"] = volume_till_now

    def __add_interval_ticks_to_market_df(self, interval):

        date = np.datetime64(self.ongoing_tick_interval_time[interval])
        date_raw = self.ongoing_tick_interval_time[interval]

        for symbol in self.ongoing_tick_updates[interval]:
            interval_tick = self.ongoing_tick_updates[interval][symbol]
            interval_tick["date"] = date

            volume = 0
            if symbol in self.prev_tick_volume[interval]:
                volume = interval_tick['total_volume'] - self.prev_tick_volume[interval][symbol]

            if date_raw.hour == 9 and date_raw.minute == 15:
                volume = interval_tick['total_volume']

            self.prev_tick_volume[interval][symbol] = interval_tick['total_volume']
            interval_tick['volume'] = volume

            for column_name in interval_tick:
                self.live_market_df_by_interval[interval].at[
                    (symbol, date), column_name
                ] = interval_tick[column_name]

        Constants.logger.info(
            f"Added Tick {self.ongoing_tick_interval_time[interval]} for interval {interval} to market_df"
        )
        self.last_updated_interval_tick[interval] = self.ongoing_tick_interval_time[
            interval
        ]
        self.processed_tick_interval_queue.put(
            (interval, self.ongoing_tick_interval_time[interval])
        )

    def last_updated_interval_tick(self, interval):
        return self.last_updated_interval_tick[interval]

    def reset_market_df_index(self):

        for interval in self.tick_intervals:
            self.live_market_df_by_interval[interval].loc[
                :, "date_index"
            ] = self.live_market_df_by_interval[interval].date
            self.live_market_df_by_interval[interval].loc[
                :, "symbol_index"
            ] = self.live_market_df_by_interval[interval].symbol
            self.live_market_df_by_interval[interval] = self.live_market_df_by_interval[
                interval
            ].set_index(["symbol_index", "date_index"])
            self.live_market_df_by_interval[interval] = self.live_market_df_by_interval[
                interval
            ].sort_values(["symbol", "date"])

    def market_feed(self, ticks):
        """Stores live ticks for the instruments added in a dataframe.
        In addition, it also persists the live data on the disk

        :param: ticks: List[QuantplayTicks]
        """

        for interval in self.tick_intervals:

            mod_value = Constants.interval_attributes[interval][
                IntervalAttribute.mod_value
            ]

            for tick in ticks:
                if not self.is_valid_live_tick(tick):
                    continue

                orders_to_modify = self.get_orders_to_modify()
                if orders_to_modify:
                    for quantplay_order in orders_to_modify:
                        self.place_or_modify_order(quantplay_order, is_new_order=False)

                last_trade_time = tick.last_trade_time
                symbol = tick.symbol
                exchange = tick.exchange
                exchange_start_time = Market.TIMINGS[exchange]["start"]

                self.exchange_symbol_last_tick[exchange][symbol] = tick

                if tick.instrument_id not in self.strategy_instruments:
                    continue

                if interval == TickInterval.day:
                    tick_interval_time = last_trade_time.replace(
                        hour=exchange_start_time[0],
                        minute=exchange_start_time[1],
                        second=0,
                    )
                else:
                    minutes_since_market_start = (
                        last_trade_time.hour - exchange_start_time[0]
                    ) * 60 + (last_trade_time.minute - exchange_start_time[1])
                    tick_number = int(minutes_since_market_start / mod_value)
                    tick_interval_time = last_trade_time.replace(
                        hour=exchange_start_time[0],
                        minute=exchange_start_time[1],
                        second=0,
                    ) + timedelta(minutes=tick_number * mod_value)

                volume_till_now = tick.volume_traded - self.pre_market_volume[symbol]
                last_price = tick.last_price

                if self.ongoing_tick_interval_time[interval] is None:
                    self.ongoing_tick_interval_time[interval] = tick_interval_time

                if (symbol in self.ongoing_tick_updates[interval]) and (
                    self.ongoing_tick_interval_time[interval] == tick_interval_time
                ):
                    self.__upsert_ongoing_tick(
                        interval,
                        symbol,
                        tick.instrument_id,
                        last_price,
                        volume_till_now,
                        initialize=False,
                    )
                else:
                    if tick_interval_time > self.ongoing_tick_interval_time[interval]:
                        self.__add_interval_ticks_to_market_df(interval)

                        self.ongoing_tick_updates[interval] = {}
                        self.__upsert_ongoing_tick(
                            interval,
                            symbol,
                            tick.instrument_id,
                            last_price,
                            volume_till_now,
                            initialize=True,
                        )
                        self.ongoing_tick_interval_time[interval] = tick_interval_time

                    if symbol not in self.ongoing_tick_updates[interval]:
                        Constants.logger.info(
                            "Found new tick for {} last_trade_time {} tick_interval_time {} volume_till_now {}".format(
                                symbol, last_trade_time, tick_interval_time, volume_till_now
                            )
                        )
                        self.__upsert_ongoing_tick(
                            interval,
                            symbol,
                            tick.instrument_id,
                            last_price,
                            volume_till_now,
                            initialize=True,
                        )

    def get_orders_to_modify(self):
        """Checks self.next_modifications_queue for any pending orders
        and returns list of orders to be modified
        """
        orders_to_modify = []
        current_time = datetime.now()
        total_qsize = self.next_modifications_queue.qsize()
        in_progress_orders = []
        elements_processed = 0

        while elements_processed < total_qsize:
            modification_time, order_id = self.next_modifications_queue.get()
            if modification_time > current_time:
                self.next_modifications_queue.put((modification_time, order_id))
                break
            elif order_id in self.in_progress_order_map:
                in_progress_orders.append((modification_time, order_id))
            elif order_id in self.last_processed_order_map:
                orders_to_modify.append(self.last_processed_order_map[order_id])
            else:
                Constants.logger.info(
                    f"not modifying order with order_id {order_id} as it is already confirmed"
                )

            elements_processed += 1

        for mod_time, order_id in in_progress_orders:
            self.next_modifications_queue.put((mod_time, order_id))

        return orders_to_modify

    def place_new_order(self, quantplay_order):
        """Translates quantplay order to broker order and calls the broker API
        for placing new orders. To be implemented in broker subclasses.

        Returns order_id, quantplay_order.
        """
        raise NotImplementedError

    def modify_exiting_order(self, quantplay_order):
        """Translates quantplay order to broker order and calls the broker API
        for modifying order. To be implemented in broker subclasses.

        Returns order_id, quantplay_order.
        """
        raise NotImplementedError

    def close_child_orders(self, tag=None):
        if not tag:
            Constants.logger.info(f"Tag must be provided for closing child orders")
            return

        if tag not in self.strategy_child_orders:
            Constants.logger.info(f"Strategy {tag} does not have any child orders")
            return

        child_orders = self.strategy_child_orders[tag]
        for order in child_orders:
            order.modify_child_order_for_closing()
            try:
                self.place_or_modify_order(order, is_new_order=False)
            except QuantplayOrderPlacementException as e:
                Constants.logger.info(
                    f"[CHILD_ORDER_MODIFICATION_FOR_CLOSING_FAILED] Failed to modify order {order} got {e}"
                )
                traceback.print_exc()

    def place_or_modify_order(self, quantplay_order, is_new_order=True):
        """Places the new order or modifies an existing order and executes it using the
        given execution algorithm.

        It stores the given order and execution algorithm against the orderId.
        """

        exchange, tradingsymbol, execution_algo = (
            quantplay_order.exchange,
            quantplay_order.tradingsymbol,
            quantplay_order.execution_algo,
        )
        instrument_id = self.exchange_symbol_to_instrument_id_map[exchange][tradingsymbol]

        if instrument_id not in self.streaming_instruments:
            self.stream_new_symbols([instrument_id])

            while tradingsymbol not in self.exchange_symbol_last_tick[exchange]:
                Constants.logger.info(
                    "Delaying order placement by 10ms as tick has not arrived for {}:{}".format(
                        exchange, tradingsymbol
                    )
                )
                time.sleep(0.1)

        self.order_placement_lock.acquire()
        try:
            if is_new_order:
                execution_algo.modify_order(
                    self.exchange_symbol_last_tick[exchange][tradingsymbol],
                    quantplay_order,
                    is_new_order=True,
                )

                order_id, quantplay_order = self.place_new_order(quantplay_order)
                Constants.logger.info(f"[NEW_ORDER_REQUEST] order_id {order_id} order {quantplay_order}")

                if quantplay_order.is_child_order == False:
                    child_order = quantplay_order.get_child_order()
                    Constants.logger.info(f"Placing child order {child_order}")
                    child_order_id, child_order = self.place_new_order(child_order)
                    Constants.logger.info(
                        f"[CHILD_ORDER_PLACED] child_order_id {child_order_id} child_order {child_order}"
                    )
                    self.in_progress_order_map[child_order_id] = child_order
                    # TODO: Map order_id to child_order_id
            else:
                execution_algo.modify_order(
                    self.exchange_symbol_last_tick[exchange][tradingsymbol],
                    quantplay_order,
                    is_new_order=False,
                )
                order_id = quantplay_order.order_id
                quantplay_order = self.modify_existing_order(quantplay_order)
                Constants.logger.info(
                    f"[MODIFY_ORDER_REQUEST] order_id {order_id} order {quantplay_order}"
                )
        except Exception as e:
            self.order_placement_lock.release()
            raise QuantplayOrderPlacementException(
                "Failed to place or modify order"
            ) from e

        self.in_progress_order_map[order_id] = quantplay_order
        self.order_placement_lock.release()

        seconds_since_modification = execution_algo.get_next_modification_time(
            self.exchange_symbol_last_tick[exchange][tradingsymbol],
            quantplay_order,
        )
        next_modification_time = quantplay_order.order_placement_time + timedelta(
            seconds=seconds_since_modification
        )
        self.next_modifications_queue.put((next_modification_time, order_id))

    def on_order_response(self, response: QuantplayExchangeResponse):
        """Processes the order response"""
        order_id = response.order_id
        message = response.message
        response_type = response.response_type
        quantity = response.quantity
        avg_price = response.average_price

        if response_type in [
            QuantplayExchangeResponseType.NEW_ORDER_CONFIRM,
            QuantplayExchangeResponseType.MOD_ORDER_CONFIRM,
            QuantplayExchangeResponseType.NEW_ORDER_REJECT,
            QuantplayExchangeResponseType.MOD_ORDER_REJECT,
        ]:

            in_progress_order = self.in_progress_order_map[order_id]
            del self.in_progress_order_map[order_id]

            if response_type in [
                QuantplayExchangeResponseType.NEW_ORDER_CONFIRM,
                QuantplayExchangeResponseType.MOD_ORDER_CONFIRM,
            ]:
                in_progress_order.transaction_status = (
                    QuantplayTransactionStatus.PROCESSED
                )
                in_progress_order.order_status = QuantplayExchangeOrderStatus.OPEN
                self.last_processed_order_map[order_id] = in_progress_order

            elif response_type == QuantplayExchangeResponseType.NEW_ORDER_REJECT:
                Constants.logger.info(f"[NEW_ORDER_REJECT] response {response}")

            elif response_type == QuantplayExchangeResponseType.MOD_ORDER_REJECT:
                Constants.logger.info(f"[MOD_ORDER_REJECT] response {response}")

        elif response_type == QuantplayExchangeResponseType.TRADE_CONFIRM:
            processed_order = self.last_processed_order_map[order_id]

            processed_order.filled_quantity += response.quantity
            if processed_order.quantity == processed_order.filled_quantity:
                processed_order.order_status = QuantplayExchangeOrderStatus.COMPLETE
                self.remove_order_from_order_maps(
                    order_id=processed_order.order_id, tag=processed_order.tag
                )

        elif response_type in [
            QuantplayExchangeResponseType.OMS_REJECT,
            QuantplayExchangeResponseType.RMS_REJECT,
        ]:
            pass
        elif response_type == QuantplayExchangeResponseType.CANCEL_ORDER_CONFIRM:
            Constants.logger.info(f"[CANCEL_ORDER_CONFIRM] response {response}")
            self.remove_order_from_order_maps(order_id=order_id)
        elif response_type == QuantplayExchangeResponseType.TRIGGER_CONFIRM:
            in_progress_order = self.in_progress_order_map[order_id]
            self.strategy_child_orders[in_progress_order.tag].append(in_progress_order)
        else:
            Constants.logger.info(f"[UNKNOWN_EXCHANGE_RESPONSE] {response}")

    def remove_order_from_order_maps(self, order_id, tag=None):

        if order_id in self.last_processed_order_map:
            del self.last_processed_order_map[order_id]

        if order_id in self.in_progress_order_map:
            Constants.logger.info(
                f"[UNEXPECTED_IN_PROGRESS_ORDER_FOUND] order_id {order_id} is complete. Deleting its unexpected presence in in_progress_order_map"
            )
            del self.in_progress_order_map[order_id]

        if tag:
            Constants.logger.info(
                f"Removing child order for tag {tag} with order_id {order_id} from strategy_child_orders map"
            )
            strategy_child_orders = self.strategy_child_orders[tag]
            strategy_child_orders_updated = [
                s_child_order
                for s_child_order in strategy_child_orders
                if s_child_order.order_id != order_id
            ]
            self.strategy_child_orders[tag] = strategy_child_orders_updated

    def get_processed_quantplay_order_by_id(self, order_id):
        """Returns the quantplay order which has been confirmed
        by the exchange for the order_id.

        Returns None if the order does not exist in memory.
        """
        self.order_placement_lock.acquire()
        order = None
        if order_id in self.last_processed_order_map:
            order = self.last_processed_order_map[order_id]
        self.order_placement_lock.release()
        return order

    def get_in_progress_quantplay_order_by_id(self, order_id):
        """Returns the quantplay order which has been requested
        to be put on the exchange for the order_id.

        Returns None if the order does not exist in memory.
        """
        self.order_placement_lock.acquire()
        order = None
        if order_id in self.in_progress_order_map:
            order = self.in_progress_order_map[order_id]
        self.order_placement_lock.release()
        return order

    def market_df(self, tick_interval):
        """Returns the market dataframe containing the live data
        for the provided interval
        """
        if tick_interval not in self.tick_intervals:
            raise Exception(
                f"Tick interval {tick_interval} has not been subscribed. Subscribed intervals: {self.tick_intervals}"
            )

        return copy.deepcopy(self.live_market_df_by_interval[tick_interval])

    def get_historical_data(self, instrument, start_date, end_date):
        """Fetches the historical data by calling the broker API.

        To Be Implemented in the actual broker's child class.
        """
        pass




import hashlib
import requests
import json
import time
import pandas as pd

from quantplay.utils.constant import Constants
from quantplay.config.qplay_config import QplayConfig
import getpass
from quantplay.broker.generics.broker import Broker
import traceback

class Motilal(Broker):
    user_id = "motilal_user_id"
    api_key = "motilal_api_key"
    password = "password"
    auth_token = "motilal_auth_token"
    two_factor_authentication = "motilal_2FA"

    url = "https://uatopenapi.motilaloswal.com/rest/login/v3/authdirectapi"
    otp_url = "https://uatopenapi.motilaloswal.com/rest/login/v3/resendotp"
    verify_otp_url = "https://uatopenapi.motilaloswal.com/rest/login/v3/verifyotp"
    ltp_utl = "https://uatopenapi.motilaloswal.com/rest/report/v1/getltpdata"
    place_order_url = "https://uatopenapi.motilaloswal.com/rest/trans/v1/placeorder"
    modify_order_url = "https://uatopenapi.motilaloswal.com/rest/trans/v2/modifyorder"
    order_book_url = "https://uatopenapi.motilaloswal.com/rest/book/v1/getorderbook"

    headers = {
        "Accept": "application/json",
        "User-Agent": "MOSL/V.1.1.0",
        "vendorinfo": "EMUM755714",
        "SourceId": "WEB",
        "MacAddress": "00:50:56:BD:F4:0B",
        "ClientLocalIp": "192.168.165.165",
        "ClientPublicIp": "106.193.137.95",
        "osname": "Ubuntu",
        "osversion": "10.0.19041",
        "devicemodel": "AHV",
        "manufacturer": "DELL",
        "productname": "Your Product Name",
        "productversion": "Your Product Version",
        "installedappid": "AppID",
        "browsername": "Chrome",
        "browserversion": "105.0"
    }

    def update_headers(self):
        Constants.logger.info("Updating headers")
        quantplay_config = QplayConfig.get_config()

        auth_token = quantplay_config['DEFAULT'][Motilal.auth_token]
        api_key = quantplay_config['DEFAULT'][Motilal.api_key]
        user_id = quantplay_config['DEFAULT'][Motilal.user_id]

        Motilal.headers['vendorinfo'] = user_id
        Motilal.headers['Authorization'] = auth_token
        Motilal.headers['ApiKey'] = api_key

    def __init__(self):
        self.symbol_scripcode_map = {"NSE": {}, "NSEFO":{}}

        try:
            if not self.validate_config():
                self.generate_token()

            self.update_headers()
            self.get_orders()
        except Exception as e:
            Constants.logger.info(traceback.print_exc())
            self.generate_token()
            self.update_headers()


        self.load_instrument()

        self.order_type_sl = "STOPLOSS"

    def load_instrument(self):
        instrument_file_FO = pd.read_csv("https://openapi.motilaloswal.com/getscripmastercsv?name=NSEFO")
        instrument_file_EQ = pd.read_csv("https://openapi.motilaloswal.com/getscripmastercsv?name=NSE")

        self.symbol_scripcode_map = pd.Series(instrument_file_EQ.scripcode.values,
                                              index=instrument_file_EQ.scripshortname).to_dict()

    def validate_config(self):
        Constants.logger.info("Validating config file")
        quantplay_config = QplayConfig.get_config()

        if quantplay_config is None:
            return False
        if Motilal.api_key not in quantplay_config['DEFAULT']:
            return False
        if Motilal.two_factor_authentication not in quantplay_config["DEFAULT"]:
            return False
        if Motilal.user_id not in quantplay_config["DEFAULT"]:
            return False
        if Motilal.auth_token not in quantplay_config["DEFAULT"]:
            return False

        Constants.logger.info("config validation successful")
        return True

    def configure(self):
        quantplay_config = QplayConfig.get_config()

        print("Enter Motilal userId:")
        user_id = input()

        print("Enter Motilal API key:")
        api_key = input()

        print("Enter Motilal 2FA:")
        two_factor_authentication = input()

        quantplay_config['DEFAULT'][Motilal.api_key] = api_key
        quantplay_config['DEFAULT'][Motilal.user_id] = user_id
        quantplay_config['DEFAULT'][Motilal.two_factor_authentication] = two_factor_authentication

        with open('{}/config'.format(QplayConfig.config_path), 'w') as configfile:
            quantplay_config.write(configfile)

    def generate_token(self):
        if not self.validate_config():
            self.configure()

        quantplay_config = QplayConfig.get_config()
        password = getpass.getpass()

        # initializing string
        str = "{}{}".format(password, quantplay_config['DEFAULT'][Motilal.api_key])
        result = hashlib.sha256(str.encode())

        data = {
            "userid": quantplay_config['DEFAULT'][Motilal.user_id],
            "password": result.hexdigest(),
            "2FA": quantplay_config['DEFAULT'][Motilal.two_factor_authentication]
        }
        Motilal.headers['ApiKey'] = quantplay_config['DEFAULT'][Motilal.api_key]
        response = requests.post(Motilal.url, headers=Motilal.headers, data=json.dumps(data))

        resp_json = response.json()
        Constants.logger.info("login response {}".format(resp_json))
        auth_token = resp_json['AuthToken']

        quantplay_config['DEFAULT'][Motilal.auth_token] = auth_token
        with open('{}/config'.format(QplayConfig.config_path), 'w') as configfile:
            quantplay_config.write(configfile)

        if "isAuthTokenVerified" in resp_json and resp_json['isAuthTokenVerified'] == "FALSE":
            print("Please enter otp")
            self.send_otp()
            otp = input()
            self.verify_otp(otp)

    def send_otp(self):
        response = requests.post(Motilal.otp_url, headers=Motilal.headers)
        Constants.logger.info(response.json())

    def verify_otp(self, otp):
        data = {
            "otp": otp
        }
        response = requests.post(Motilal.verify_otp_url, headers=Motilal.headers, data=json.dumps(data))
        Constants.logger.info(response.json())

    def get_ltp(self, exchange=None, tradingsymbol=None):
        data = {
            "userid": "EMUM755714",
            "exchange": exchange,
            "scripcode": self.symbol_scripcode_map[tradingsymbol]
        }

        response = requests.post(Motilal.ltp_utl, headers=Motilal.headers, data=json.dumps(data))
        Constants.logger.info("Get LTP response {}".format(response.json()))
        return response.json()["data"]["ltp"]/100.0

    def get_orders(self):
        response = (requests.post(Motilal.order_book_url, headers=Motilal.headers)).json()
        if response["status"] == "ERROR":
            Constants.logger.info("Error while fetching order book [{}]".format(response["message"]))
            raise Exception(response["message"])
        orders = response["data"]
        return orders

    def modify_orders_till_complete(self, orders_placed):
        modification_count = {}
        while 1:
            time.sleep(10)
            orders = pd.DataFrame(self.get_orders())
            orders = orders[orders.uniqueorderid.isin(orders_placed)]

            #TODO add complete, cancelled, rejected status here
            orders = orders[~orders.status.isin(["Error"])]

            if len(orders) == 0:
                Constants.logger.info("ALL orders have be completed")
                break

            orders = orders.to_dict('records')
            for order in orders:
                order_id = order['uniqueorderid']

                ltp = self.get_ltp(order['exchange'], order['symbol'])
                order['price'] = ltp
                self.modify_order(order)

                if order_id not in modification_count:
                    modification_count[order_id] = 1
                else:
                    modification_count[order_id] += 1

                time.sleep(.1)

                if modification_count[order_id] > 5:
                    order['ordertype'] = "MARKET"
                    order['price'] = 0
                    Constants.logger.info("Placing MARKET order [{}]".format(order))
                    self.modify_order(order)

    def modify_order(self, order):
        data = {
          "uniqueorderid ": order['uniqueorderid'],
          "newordertype": order['ordertype'],
          "neworderduration": order['orderduration'],
          " newquantityinlot ": order['totalqtyremaining'],
          "newdisclosedquantity": 0,
          "newprice": order['price'],
          "newtriggerprice": order['triggerprice'],
          "qtytradedtoday": order['qtytradedtoday']
        }

        try:
            Constants.logger.info("Modifying order [{}] new price [{}]".format(order['uniqueorderid'], order['price']))
            response = requests.post(Motilal.modify_order_url, headers=Motilal.headers, data=json.dumps(data)).json()
        except Exception as e:
            exception_message = "OrderModificationFailed for {} failed with exception {}".format(order['uniqueorderid'], e)
            Constants.logger.error("{}".format(exception_message))

    def place_order(self, tradingsymbol=None, exchange=None, quantity=None, order_type=None, transaction_type=None,
                    tag=None, product=None, price=None, trigger_price=None):
        data = {
            "exchange": exchange,
            "symboltoken": self.symbol_scripcode_map[tradingsymbol],
            "buyorsell": transaction_type,
            "ordertype": order_type,
            "producttype": product,
            "orderduration": "DAY",
            "price": price,
            "triggerprice": trigger_price,
            "quantityinlot": quantity,
            "disclosedquantity": 0,
            "amoorder": "N",
            "algoid": "",
            "tag": tag
        }
        try:
            Constants.logger.info("Placing order {}".format(json.dumps(data)))
            response = requests.post(Motilal.place_order_url, headers=Motilal.headers, data=json.dumps(data)).json()
            if response['status'] == "ERROR":
                raise Exception(response['message'])
            return response['uniqueorderid']
        except Exception as e:
            exception_message = "Order placement failed with error [{}]".format(str(e))
            print(exception_message)


from quantplay.strategy.base import QuantplayAlgorithm
from quantplay.utils.constant import TickInterval
from quantplay.service import market
from quantplay.order_execution.mean_price import MeanPriceExecutionAlgo
import numpy as np
import pandas as pd


class Musk(QuantplayAlgorithm):
    def __init__(self):

        # Mandatory Attributes
        self.interval = TickInterval.minute
        self.entry_time = "09:29"
        self.exit_time = "15:15"
        self.strategy_trigger_times = [self.entry_time]
        self.exchange_to_trade_on = "NFO"
        self.option_nearest_expiry_offset = 0
        self.option_chain_depth = 0
        self.backtest_after_date = "2021-01-01"
        self.backtest_before_date = "2022-02-25"
        self.stream_symbols_by_security_type = {"EQ": ["NIFTY BANK"]}
        self.columns_for_uuid = ["date", "symbol"]
        self.exact_number_of_orders_per_uuid = 2
        self.strategy_type = "intraday"
        self.strategy_tag = "musk"
        self.execution_algo = MeanPriceExecutionAlgo(7)

        self.data_required_for_days = 20

        super(Musk, self).__init__()

    def get_trades(self, market_data):
        equity_data = market_data[(market_data.security_type == "EQ") & (market_data.symbol != "SBIN")]

        unique_equity_symbols = list(equity_data.symbol.unique())
        assert len(unique_equity_symbols) == 2
        assert set(unique_equity_symbols) == {"NIFTY 50", "NIFTY BANK"}

        trades = market.get_trades(equity_data, self.entry_time)

        trades = self.add_expiry(trades, security_type="OPT")
        trades.to_csv("/tmp/{}.csv".format(self.strategy_tag))
        trades = trades[trades.date.dt.year >= 2019]

        trades = trades[trades.strike_gap > 0]
        trades.loc[:, "atm_price"] = (
            round(trades.close / trades.strike_gap) * trades.strike_gap
        )
        trades.loc[:, "atm_price"] = trades.atm_price.astype(int)

        pe_trades = market.option_symbol(
            trades, price_column="atm_price", option_type="PE"
        )
        ce_trades = market.option_symbol(
            trades, price_column="atm_price", option_type="CE"
        )

        trades = pd.concat([pe_trades, ce_trades])

        trades.loc[:, "day_of_week"] = trades.date.dt.day_name()
        trades = trades[trades.day_of_week.isin(["Monday", "Tuesday", "Wednesday", "Thursday"])]

        trades.loc[:, "transaction_type"] = "SELL"
        trades.loc[:, "stoploss"] = np.where(trades.symbol == "NIFTY 50", 0.5, 0.8)
        trades.loc[:, "quantity"] = np.where(trades.symbol == "NIFTY 50", 100, 50)

        trades = self.filter_uuids_not_matching_count(trades)

        return trades


